%{

// -------------------------------------------------------------------
// TDL.y
//
// This file is a standard input to the compiler compiler bison 
// (which is a variant of yacc). It provides a formal grammar for 
// the IEEE 1451.4 Template Description Language (TDL).  To use 
// this file, you must first create the file lexyy.cpp using the 
// file tdl.l (which contains the token definitions).  You then 
// run this file through bison by executing the command line:
//
//		bison -d tdl.y
//
// After executing this command do:
//
//		rename tdl_tab.c tdl_tab.cpp.  
//
// You can then compile the main TDL compiler using a C++ compiler.
//
// Note that the stack generated by bison in tdl_tab.c may not 
// be big enough.  The sympton of this is that the parser will appear to 
// terminate properly but will not have completed parsing the input 
// file. It might be necessary, inside tdl_tab.cpp, to change
//
//	#define YYINITDEPTH 200
//
// to
//
//	#define YYINITDEPTH 500
//
// One aspect of the design of this grammar that could probably be
// done better is the use of end_of_line.  The need to worry about this
// as something other than white space is a consequence of both the
// comment and abstract commands that use end of line as a delimiter.
//
// Updated 10 September 2003
//
// -------------------------------------------------------------------

#include "tdlparser.h"  // Main header file for TDL compiler.

// -------------------------------------------------------------------
// lex and yacc stuff
// -------------------------------------------------------------------

extern int yyerror (char*);
extern int yylex ();

// -------------------------------------------------------------------
// debug variables
// -------------------------------------------------------------------

extern int statement_count;
extern void print_statement_num ();

// -------------------------------------------------------------------
// semantic action routine prototypes
// -------------------------------------------------------------------

extern void initialize_parser ();
extern void finalize_parser ();
extern void initialize_template ();
extern void finalize_template ();
extern void process_abstract_statement();
extern void process_align ();
extern void process_case ();			
extern void process_case_end ();			
extern void process_comment();
extern void process_enumerate ();			
extern void process_end_enumerate ();			
extern void process_enumerate_item ();			
extern void process_physical_unit ();
extern void process_property_definition ();
extern void process_selectcase ();
extern void process_endselect ();
extern void process_spacing ();
extern void process_structarray_statement();
extern void process_endstructarray_statement();
extern void process_subproperty ();
extern void process_template_statement();
extern void process_tdl_version_number ();
extern void process_ugid ();
extern void process_validation_keycode ();

%}

// -------------------------------------------------------------------
// All tokens sent on to the action routines are captured
// as strings.  It is up to the action routines to convert
// to appropriate data types.
// -------------------------------------------------------------------

%union {
  double real;
  char* str;
  int integer;
}

%token <str> REAL 
%token <str> INTEGER
%token <str> QUOTED_STRING
%token <str> HEX_INTEGER
%token <str> BINARY_INTEGER
%token <str> PROPERTYNAME
%token <str> VARIABLE_NAME
%token <str> TEMPLATE
%token <str> COMMENT
%token <str> ABSTRACT

%type <str> size boundary template_statement
%type <str> numeric case_bit_value description
%type <str> selectcase_number_of_bits startvalue tolerance 
%type <str> manufacturer_code id_number_of_bits template_id

%token VALIDATION_KEYCODE TDL_VERSION_NUMBER SPACING 
%token ENDTEMPLATE SELECTCASE ENDSELECT CASE ENDCASE 
%token STRUCTARRAY ENDSTRUCTARRAY
%token ID CAL USR ALIGN UGID
%token UNINT CHR5 DATE CONRELRES CONRES BITBIN
%token STRING5 STRING7 STRING16
%token ASCII SINGLE UNICOD // UNICODE conflicts with a C++ variable
%token END_OF_LINE
%token DEFAULT INITIALIZE READWRITE CTRLFUNCTIONMASK 
%token FUNCTIONTYPE FUNCTION 
%token ENUMERATE PHYSICAL_UNIT

%%

file : between_templates_list template_list optional_eol {}

template_list : {initialize_parser ();} 
                template between_templates_list 
                template_list 
        | VALIDATION_KEYCODE_statement
        {process_validation_keycode (); finalize_parser ();} 
between_templates_list : | between_templates between_templates_list
between_templates : comment_statement | eol

template : {initialize_template ();} template_statement statement_list 
           ENDTEMPLATE {finalize_template ();}
template_statement : TEMPLATE manufacturer_code comma 
            id_number_of_bits comma template_id ','
            optional_description {process_template_statement();}  
            eol optional_eol

manufacturer_code : INTEGER {strcpy (manufacturer_code, $1);}
id_number_of_bits : INTEGER {strcpy (id_number_of_bits, $1);}
template_id : INTEGER {strcpy (template_id, $1);}
optional_description : {description = NULL;} | description {}

statement_list : | statement optional_eol statement_list
statement : abstract_statement {process_abstract_statement();} eol
          | align_statement {process_align ();}
          | comment_statement
          | enumerate_statement
          | physical_unit_statement
          | property_definition_command 
            {process_property_definition ();}
          | selectcase_statement
          | SPACING {process_spacing ();}
          | structarray_statement
          | TDL_VERSION_NUMBER_statement
            {process_tdl_version_number ();}
          | UGID_statement {process_ugid ();}

abstract_statement  : ABSTRACT {abstract = $1;}
align_statement	    : ALIGN boundary 
comment_statement   : COMMENT {comment = $1; process_comment();}

enumerate_statement	: ENUMERATE VARIABLE_NAME {enum_name = $2; 
                      process_enumerate();} enumerate_list 
                      {process_end_enumerate();}
enumerate_list      : comma enumerate_item optional_enumerate_item 
					
optional_enumerate_item : | comma enumerate_item optional_enumerate_item
enumerate_item      : QUOTED_STRING {variable_name = $1; 
                      strcpy (numeric, ""); process_enumerate_item ();} 
                    | VARIABLE_NAME {variable_name = $1; 
                      strcpy (numeric, ""); process_enumerate_item ();} 
                    | numeric {variable_name = NULL; 
                      strcpy (numeric, $1); process_enumerate_item ();} 

physical_unit_statement : PHYSICAL_UNIT QUOTED_STRING comma dot2_unit 
                          {variable_name = $2; process_physical_unit ();}

property_definition_command : property comma optional_prop_description comma
            accesslevel comma size comma datatype comma format ','
            physical_unit optional_assignment

property : PROPERTYNAME {propertyname = $1;} optional_subproperty {}

optional_subproperty : {strcpy(subproperty, ""); register_mask = NULL;}
                     | '[' subproperty ']' 
                       {process_subproperty();} optional_eol
subproperty : DEFAULT {strcpy(subproperty, "DEFAULT"); }
            | INITIALIZE {strcpy(subproperty, "INITIALIZE");}
            | READWRITE {strcpy(subproperty, "READWRITE");} 
            | CTRLFUNCTIONMASK {strcpy(subproperty, "CTRLFUNCTIONMASK");}
            | QUOTED_STRING {register_mask = $1;}
            | FUNCTIONTYPE {strcpy(subproperty, "FUNCTIONTYPE");}
            | FUNCTION {strcpy(subproperty, "FUNCTION");}

optional_prop_description : {description = NULL; 
                             function_propertyname = NULL;
                             function_register_mask = NULL} 
            | description {function_propertyname = NULL;
                           function_register_mask = NULL} 
            | PROPERTYNAME {description = NULL; function_propertyname = $1;} 
              '[' QUOTED_STRING {function_register_mask = $4;} ']'

accesslevel : {strcpy(access_level,"");} 
            | ID {strcpy(access_level,"ID");}
            | CAL {strcpy(access_level,"CAL");}
            | USR {strcpy(access_level, "USR");}

datatype : UNINT    {strcpy(datatype, "UNINT");}
         | CHR5	    {strcpy(datatype, "CHR5");}
         | DATE	    {strcpy(datatype, "DATE");}
         | CONRELRES comma startvalue comma tolerance 
                    {strcpy(datatype, "ConRelRes");}
         | CONRES comma startvalue comma tolerance 
                    {strcpy(datatype, "ConRes");}
         | ASCII    {strcpy(datatype, "ASCII");}
         | STRING5  {strcpy(datatype, "String5");}
         | STRING7  {strcpy(datatype, "String7");}
         | STRING16 {strcpy(datatype, "String16");}
         | SINGLE   {strcpy(datatype, "Single");}
         | UNICOD   {strcpy(datatype, "Unicode");}
         | BITBIN   {strcpy(datatype, "BitBin");}
         | VARIABLE_NAME {enum_name = $1; 
                     strcpy(datatype, "Enumeration");}

format : {format = NULL;} | QUOTED_STRING {format = $1;} 

physical_unit : {physical_unit = NULL;} 
              | QUOTED_STRING {physical_unit=$1;}
              | dot2_unit {physical_unit = NULL}
dot2_unit : left_paren numeric {strcpy(dot2_1, $2);}
            comma numeric {strcpy(dot2_2, $5);}
            comma numeric {strcpy(dot2_3, $8);}
            comma numeric {strcpy(dot2_4, $11);}
            comma numeric {strcpy(dot2_5, $14);}
            comma numeric {strcpy(dot2_6, $17);}
            comma numeric {strcpy(dot2_7, $20);}
            comma numeric {strcpy(dot2_8, $23);}
            comma numeric {strcpy(dot2_9, $26);}
            comma numeric {strcpy(dot2_10, $29);}
            comma numeric {strcpy(dot2_11, $32);}
            comma numeric {strcpy(dot2_12, $35);}
            right_paren 
		
optional_assignment : {assignment_found = false;}
                    | equals {assignment_found = true; hex_integer=NULL;
                      binary_integer = NULL; variable_name = NULL;
                      assignment_string = NULL; strcpy(numeric, "");} 
                      assignment
assignment : HEX_INTEGER {hex_integer = $1;}
           | BINARY_INTEGER {binary_integer = $1;}
           | VARIABLE_NAME {variable_name = $1;}
           | numeric {strcpy (numeric,$1);}
           | QUOTED_STRING {assignment_string = $1;}

selectcase_statement : SELECTCASE description comma accesslevel comma 
                       selectcase_number_of_bits 
                       {process_selectcase ();} optional_eol
                       case_statement_list ENDSELECT 
                       {process_endselect ();}
case_statement_list : | case_statement case_statement_list
case_statement : CASE description comma case_bit_value 
                 {process_case ();} optional_eol statement_list 
                 ENDCASE {process_case_end();} optional_eol

structarray_statement : STRUCTARRAY VARIABLE_NAME {variable_name = $2;} 
                        comma description comma accesslevel comma size 
                        {process_structarray_statement ();} optional_eol
                        struct_block ENDSTRUCTARRAY 
                        {process_endstructarray_statement ();}
struct_block : | struct_block_statement struct_block
struct_block_statement : property_definition_command optional_eol 
                         {process_property_definition ();}
                       | structarray_statement optional_eol

TDL_VERSION_NUMBER_statement : TDL_VERSION_NUMBER INTEGER 
                               {strcpy (tdl_version_number, $2);}
UGID_statement : UGID QUOTED_STRING {ugid=$2;}
                 comma description

VALIDATION_KEYCODE_statement : VALIDATION_KEYCODE INTEGER 
                               {strcpy (validation_keycode, $2);}

description : QUOTED_STRING {description=$1;}

case_bit_value : INTEGER {strcpy (case_bit_value, $1);}
selectcase_number_of_bits : INTEGER 
                            {strcpy (selectcase_number_of_bits, $1);}
size : INTEGER {strcpy (size,$1);}
boundary : INTEGER {strcpy (boundary,$1);}

startvalue : numeric {strcpy (startvalue,$1);}
tolerance : numeric {strcpy (tolerance,$1);}

numeric : REAL | INTEGER

comma       : ',' optional_eol
equals      : '=' optional_eol
left_paren  : '(' optional_eol
right_paren : optional_eol ')' 

eol : END_OF_LINE {statement_count++; print_statement_num();}
optional_eol : | eol optional_eol

%%
